#define __SP_QUICK_ALIGN_C__

/*
 * Object align dialog
 *
 * Authors:
 *   Frank Felfe <innerspace@iname.com>
 *   Lauris Kaplinski <lauris@kaplinski.com>
 *
 * Copyright (C) 1999-2002 Authors
 *
 * Released under GNU GPL, read the file 'COPYING' for more information
 */

#include <config.h>

#include <math.h>
#include <glib.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkwindow.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtktable.h>
#include <gtk/gtkoptionmenu.h>
#include <gtk/gtkmenuitem.h>
#include <gtk/gtkimage.h>

#include "helper/sp-intl.h"
#include "widgets/button.h"
#include "sodipodi.h"
#include "document.h"
#include "desktop-handles.h"
#include "sp-item-transform.h"
#include "selection.h"

#include "align.h"

/*
 * handler functions for quick align dialog
 *
 * todo: dialog with more aligns
 * - more aligns (30 % left from center ...)
 * - aligns for nodes
 *
 */ 

enum {
	SP_ALIGN_LAST,
	SP_ALIGN_FIRST,
	SP_ALIGN_BIGGEST,
	SP_ALIGN_SMALLEST,
	SP_ALIGN_PAGE,
	SP_ALIGN_DRAWING,
	SP_ALIGN_SELECTION,
};

enum {
	SP_ALIGN_TOP_IN,
	SP_ALIGN_TOP_OUT,
	SP_ALIGN_RIGHT_IN,
	SP_ALIGN_RIGHT_OUT,
	SP_ALIGN_BOTTOM_IN,
	SP_ALIGN_BOTTOM_OUT,
	SP_ALIGN_LEFT_IN,
	SP_ALIGN_LEFT_OUT,
	SP_ALIGN_CENTER_HOR,
	SP_ALIGN_CENTER_VER,
};

static const unsigned char aligns[10][8] = {
	{0, 0, 0, 2, 0, 0, 0, 2},
	{0, 0, 0, 2, 0, 0, 2, 0},
	{0, 2, 0, 0, 0, 2, 0, 0},
	{0, 2, 0, 0, 2, 0, 0, 0},
	{0, 0, 2, 0, 0, 0, 2, 0},
	{0, 0, 2, 0, 0, 0, 0, 2},
	{2, 0, 0, 0, 2, 0, 0, 0},
	{2, 0, 0, 0, 0, 2, 0, 0},
	{1, 1, 0, 0, 1, 1, 0, 0},
	{0, 0, 1, 1, 0, 0, 1, 1}
};

void sp_align_arrange_clicked (GtkWidget *widget, const unsigned char *aligns);

void sp_quick_align_dialog_close (void);

static GtkWidget * create_base_menu (void);
static void set_base (GtkMenuItem * menuitem, gpointer data);
static SPItem * sp_quick_align_find_master (const GSList * slist, gboolean horizontal);

static GtkWidget *dlg = NULL;

static unsigned int base = SP_ALIGN_LAST;

static int
sp_quick_align_dialog_delete (void)
{
	if (GTK_WIDGET_VISIBLE (dlg)) gtk_widget_hide (dlg);

	return TRUE;
}

static void
sp_align_add_button (GtkWidget *t, int col, int row, GCallback handler, gconstpointer data, const unsigned char *px, const unsigned char *tip)
{
	GtkWidget *b;

	b = sp_button_new (24, px, tip);
	gtk_widget_show (b);
	if (handler) g_signal_connect (G_OBJECT (b), "clicked", handler, (gpointer) data);
	gtk_table_attach (GTK_TABLE (t), b, col, col + 1, row, row + 1, 0, 0, 0, 0);
}

void
sp_quick_align_dialog (void)
{
	if (!dlg) {
		GtkWidget *vb, *om, *t;

		dlg = gtk_window_new (GTK_WINDOW_TOPLEVEL);
		gtk_window_set_title (GTK_WINDOW (dlg), _("Align objects"));
		g_signal_connect (G_OBJECT (dlg), "delete_event", G_CALLBACK (sp_quick_align_dialog_delete), NULL);

		vb = gtk_vbox_new (FALSE, 4);
		gtk_widget_show (vb);
		gtk_container_set_border_width (GTK_CONTAINER (vb), 4);
		gtk_container_add (GTK_CONTAINER (dlg), vb);

		om = gtk_option_menu_new ();
		gtk_widget_show (om);
		gtk_box_pack_start (GTK_BOX (vb), om, FALSE, FALSE, 0);
		gtk_option_menu_set_menu (GTK_OPTION_MENU (om), create_base_menu ());

		t = gtk_table_new (2, 5, TRUE);
		gtk_widget_show (t);
		gtk_box_pack_start (GTK_BOX (vb), t, FALSE, FALSE, 0);

		sp_align_add_button (t, 0, 0, G_CALLBACK (sp_align_arrange_clicked), aligns[SP_ALIGN_LEFT_OUT], "al_left_out",
				     _("Right side of aligned objects to left side of anchor"));
		sp_align_add_button (t, 1, 0, G_CALLBACK (sp_align_arrange_clicked), aligns[SP_ALIGN_LEFT_IN], "al_left_in",
				     _("Left side of aligned objects to left side of anchor"));
		sp_align_add_button (t, 2, 0, G_CALLBACK (sp_align_arrange_clicked), aligns[SP_ALIGN_CENTER_HOR], "al_center_hor",
				     _("Center horizontally"));
		sp_align_add_button (t, 3, 0, G_CALLBACK (sp_align_arrange_clicked), aligns[SP_ALIGN_RIGHT_IN], "al_right_in",
				     _("Right side of aligned objects to right side of anchor"));
		sp_align_add_button (t, 4, 0, G_CALLBACK (sp_align_arrange_clicked), aligns[SP_ALIGN_RIGHT_OUT], "al_right_out",
				     _("Left side of aligned objects to right side of anchor"));

		sp_align_add_button (t, 0, 1, G_CALLBACK (sp_align_arrange_clicked), aligns[SP_ALIGN_TOP_OUT], "al_top_out",
				     _("Bottom of aligned objects to top of anchor"));
		sp_align_add_button (t, 1, 1, G_CALLBACK (sp_align_arrange_clicked), aligns[SP_ALIGN_TOP_IN], "al_top_in",
				     _("Top of aligned objects to top of anchor"));
		sp_align_add_button (t, 2, 1, G_CALLBACK (sp_align_arrange_clicked), aligns[SP_ALIGN_CENTER_VER], "al_center_ver",
				     _("Center vertically"));
		sp_align_add_button (t, 3, 1, G_CALLBACK (sp_align_arrange_clicked), aligns[SP_ALIGN_BOTTOM_IN], "al_bottom_in",
				     _("Bottom of aligned objects to bottom of anchor"));
		sp_align_add_button (t, 4, 1, G_CALLBACK (sp_align_arrange_clicked), aligns[SP_ALIGN_BOTTOM_OUT], "al_bottom_out",
				     _("Top of aligned objects to top of anchor"));
	}

	if (!GTK_WIDGET_VISIBLE (dlg)) gtk_widget_show (dlg);
}

void
sp_quick_align_dialog_close (void)
{
	g_assert (dlg != NULL);

	if (GTK_WIDGET_VISIBLE (dlg)) gtk_widget_hide (dlg);
}

static void
sp_align_add_menuitem (GtkWidget *menu, const unsigned char *label, GCallback handler, int value)
{
	GtkWidget *menuitem;

	menuitem = gtk_menu_item_new_with_label (label);
	gtk_widget_show (menuitem);
	if (handler) g_signal_connect (G_OBJECT (menuitem), "activate", handler, GINT_TO_POINTER (value));
	gtk_menu_append (GTK_MENU (menu), menuitem);
}

static GtkWidget *
create_base_menu (void)
{
	GtkWidget *menu;

	menu = gtk_menu_new ();

	sp_align_add_menuitem (menu, _("Last selected"), G_CALLBACK (set_base), SP_ALIGN_LAST);
	sp_align_add_menuitem (menu, _("First selected"), G_CALLBACK (set_base), SP_ALIGN_FIRST);
	sp_align_add_menuitem (menu, _("Biggest item"), G_CALLBACK (set_base), SP_ALIGN_BIGGEST);
	sp_align_add_menuitem (menu, _("Smallest item"), G_CALLBACK (set_base), SP_ALIGN_SMALLEST);
	sp_align_add_menuitem (menu, _("Page"), G_CALLBACK (set_base), SP_ALIGN_PAGE);
	sp_align_add_menuitem (menu, _("Drawing"), G_CALLBACK (set_base), SP_ALIGN_DRAWING);
	sp_align_add_menuitem (menu, _("Selection"), G_CALLBACK (set_base), SP_ALIGN_SELECTION);

	gtk_widget_show (menu);

	return menu;
}

static void
set_base (GtkMenuItem *menuitem, gpointer data)
{
	base = GPOINTER_TO_UINT (data);
}


void
sp_align_arrange_clicked (GtkWidget *widget, const unsigned char *aligns)
{
	float mx0, mx1, my0, my1;
	float sx0, sx1, sy0, sy1;
	SPDesktop * desktop;
	SPSelection * selection;
	GSList * slist;
	SPItem * master, * item;
	NRRectF b;
	NRPointF mp, sp;
	GSList * l;
	gboolean changed;

	mx0 = 0.5 * aligns[0];
	mx1 = 0.5 * aligns[1];
	my0 = 0.5 * aligns[2];
	my1 = 0.5 * aligns[3];
	sx0 = 0.5 * aligns[4];
	sx1 = 0.5 * aligns[5];
	sy0 = 0.5 * aligns[6];
	sy1 = 0.5 * aligns[7];

	desktop = SP_ACTIVE_DESKTOP;
	if (!desktop) return;

	selection = SP_DT_SELECTION (desktop);
	slist = (GSList *) sp_selection_item_list (selection);
	if (!slist) return;

	switch (base) {
	case SP_ALIGN_LAST:
	case SP_ALIGN_FIRST:
	case SP_ALIGN_BIGGEST:
	case SP_ALIGN_SMALLEST:
		if (!slist->next) return;
		slist = g_slist_copy (slist);
		master = sp_quick_align_find_master (slist, (mx0 != 0.0) || (mx1 != 0.0));
		slist = g_slist_remove (slist, master);
		sp_item_bbox_desktop (master, &b);
		mp.x = mx0 * b.x0 + mx1 * b.x1;
		mp.y = my0 * b.y0 + my1 * b.y1;
		break;
	case SP_ALIGN_PAGE:
		slist = g_slist_copy (slist);
		mp.x = mx1 * sp_document_width (SP_DT_DOCUMENT (desktop));
		mp.y = my1 * sp_document_height (SP_DT_DOCUMENT (desktop));
		break;
	case SP_ALIGN_DRAWING:
		slist = g_slist_copy (slist);
		sp_item_bbox_desktop ((SPItem *) sp_document_root (SP_DT_DOCUMENT (desktop)), &b);
		mp.x = mx0 * b.x0 + mx1 * b.x1;
		mp.y = my0 * b.y0 + my1 * b.y1;
		break;
	case SP_ALIGN_SELECTION:
		slist = g_slist_copy (slist);
		sp_selection_bbox (selection, &b);
		mp.x = mx0 * b.x0 + mx1 * b.x1;
		mp.y = my0 * b.y0 + my1 * b.y1;
		break;
	default:
		g_assert_not_reached ();
		break;
	};

	changed = FALSE;

	for (l = slist; l != NULL; l = l->next) {
		item = (SPItem *) l->data;
		sp_item_bbox_desktop (item, &b);
		sp.x = sx0 * b.x0 + sx1 * b.x1;
		sp.y = sy0 * b.y0 + sy1 * b.y1;

		if ((fabs (mp.x - sp.x) > 1e-9) || (fabs (mp.y - sp.y) > 1e-9)) {
			sp_item_move_rel (item, mp.x - sp.x, mp.y - sp.y);
			changed = TRUE;
		}
	}

	g_slist_free (slist);

	if (changed) {
		sp_selection_changed (selection);
		sp_document_done (SP_DT_DOCUMENT (desktop));
	}
}

static SPItem *
sp_quick_align_find_master (const GSList *slist, gboolean horizontal)
{
	NRRectF b;
	const GSList * l;
	SPItem * master, * item;
	gdouble dim, max;

	master = NULL;

	switch (base) {
	case SP_ALIGN_LAST:
		return (SPItem *) slist->data;
		break;
	case SP_ALIGN_FIRST: 
		return (SPItem *) g_slist_last ((GSList *) slist)->data;
		break;
	case SP_ALIGN_BIGGEST:
		max = -1e18;
		for (l = slist; l != NULL; l = l->next) {
			item = (SPItem *) l->data;
			sp_item_bbox_desktop (item, &b);
			if (horizontal) {
				dim = b.x1 - b.x0; 
			} else {
				dim = b.y1 - b.y0;
			}
			if (dim > max) {
				max = dim;
				master = item;
			}
		}
		return master;
		break;
	case SP_ALIGN_SMALLEST:
		max = 1e18;
		for (l = slist; l != NULL; l = l->next) {
			item = (SPItem *) l->data;
			sp_item_bbox_desktop (item, &b);
			if (horizontal) {
				dim = b.x1 - b.x0;
			} else {
				dim = b.y1 - b.y0;
			}
			if (dim < max) {
				max = dim;
				master = item;
			}
		}
		return master;
		break;
	default:
		g_assert_not_reached ();
		break;
	}

	return NULL;
}





